//LPacker beta

//Copyright 2008 Giannis (John) Chatzikonstantinou
//http://strukturl.blogspot.com
//http://xconst.freeshell.org

// *This file is part of LPacker*
/**
    LPacker is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    LPacker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with LPacker.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * 
 * This is the actual data structure that becomes the
 * building block of the tree.
 * Its basic properties are its dimensions, its children and its type (leaf, free)
 * 
 * It contains facilities for:
 * - Automatically finding areas denoted as free to insert a spcecified element(returns false otherwise)
 * - Recalculating the dimensions of all the children elements based on the leaf elements(real dimensions)
 * - Recalculating real free space after any transformation
 * - Swapping its children
 */

// X and Y variables do not participate in the packing process (the creation of the data tree).
// they are only used later on for speeding up the drawing process and interaction

class lrect {
  float W, H, A; // dimensions and area. WARNING: Area shows TRUE area occupied, as
  // in: Area occupied by lr1+lr2, or by self if leaf, NOT by W*H (which is different)
  float X=0,Y=0; // These are only to be used as helper variables to reduce calculations. 
  // The actual packing process ignores them.
  boolean direction; //0(false)=row cut, 1(true)=column cut
  boolean leaf; //Denotes a leaf in the tree structure (dimensions are important)
  boolean free; //Denotes a free leaf in the tree structure (dimensions are important)
  boolean capfront=false;
  boolean caprear=false;
  lrect lr1, lr2;
  lrect(lrect lr1in, lrect lr2in, boolean sdirin) { // 0=row cut, 1=column cut
    lr1=lr1in;
    lr2=lr2in;
    A=lr1.A+lr2.A;
    leaf=false;
    free=false;
    if (sdirin) {
      direction=true;
      W=lr1.W+lr2.W;
      H=max(lr1.H,lr2.H);
    } 
    else {
      direction=false;
      W=max(lr1.W,lr2.W);
      H=lr1.H+lr2.H;
    }
  }
  lrect(float Win, float Hin, boolean freein) {
    W=Win;
    H=Hin;
    A=W*H;
    if (!freein) {
      leaf=true;
      free=false;
    } 
    else {
      leaf=false;
      free=true;
    }
  }
  lrect() {
  }
  lrect(float Xin, float Yin, float Win, float Hin, boolean freein) {
    X=Xin;
    Y=Yin;
    W=Win;
    H=Hin;
    A=W*H;
    if (!freein) {
      leaf=true;
      free=false;
    } 
    else {
      leaf=false;
      free=true;
    }
  }
  boolean insert(lrect lrin) {
    lrect lr=lrin;
    if (!leaf && !free) {
      boolean newlr = lr1.insert(lr);
      if (newlr) {
        return (true);
      }
      return(lr2.insert(lr));
    } 
    else {
      if (!free) {
        return (false);
      }
      if (lr.W>W || lr.H>H) {
        return (false);
      }
      if ((int)lr.W==(int)W && (int)lr.H==(int)H) {
        free=false;
        leaf=true;
        return(true);
      }
      float dw=W-lr.W;
      float dh=H-lr.H;
// Needs optimization. To create only nonzero free space.
      if (dw>dh) {
        free=false;
        leaf=false;
        direction=true;
        lr1=new lrect(lr,new lrect(lr.W,H-lr.H,true),false);
        lr2=new lrect(W-lr.W,H,true);
      } 
      else {
        free=false;
        leaf=false;
        direction=false;
        lr1=new lrect(lr,new lrect(W-lr.W,lr.H,true),true);
        lr2=new lrect(W,H-lr.H,true);
      }
      return(true);
    }
  }
  // Applies to self all object properties from the argument object
  void clonefrom(lrect lrin) {
    lrect lr=lrin;
    W=lr.W;
    H=lr.H;
    A=lr.A;
    direction=lr.direction;
    leaf=lr.leaf;
    free=lr.free;
    if (!lr.free && !lr.leaf) {
      lr1=lr.lr1;
      lr2=lr.lr2;
    }
  }
  // Updates the dimensions (W,H,A) of the structure, based on the
  // dimensions of the leaf rectangles. (real dimensions)
  // Items affected are non-leaf non-free lrects and leaf lrects.
  // During this process, free areas are not taken into account. As a result,
  // free areas will not conform to the structure's dimensions, unless updatefree()
  // is called (look below v)
  // Does not affect positions (X,Y)
  void updatedims() {
    if (!free && !leaf) {
      lr1.updatedims();
      lr2.updatedims();
      W=0; 
      H=0; 
      A=0;
      if (direction) {//column cut
        // Calculate dimensions
        if (!lr1.free) {
          W+=lr1.W;
          H=lr1.H;
          A+=lr1.A;
        }
        if (!lr2.free) {
          W+=lr2.W;
          H=max(H,lr2.H);
          A+=lr2.A;
        }
      } 
      else {//row cut
        // Calculate dimensions
        W=0; 
        H=0; 
        A=0;
        if (!lr1.free) {
          H+=lr1.H;
          W=lr1.W;
          A+=lr1.A;
        }
        if (!lr2.free) {
          H+=lr2.H;
          W=max(W,lr2.W);
          A+=lr2.A;
        }
      }
    }
  }
  //Swaps children.
  void mirror() {
    lrect tlr=lr1;
    lr1=lr2;
    lr2=tlr;
  }
}

// This is essentially a line object with a reference to an lrect (the one above^), 
// in order to specify slicing lines to draw and interact with
class lline {
  float X1,Y1,X2,Y2;
  lrect lr;
  lline(float X1in, float  Y1in, float  X2in, float  Y2in, lrect lrin) {
    X1=X1in;
    X2=X2in;
    Y1=Y1in;
    Y2=Y2in;
    lr=lrin;
  }
}
// Completely removes all free areas
void removefree(lrect lrin) {
  lrect lr=lrin;
  if (!lr.free && !lr.leaf) {
    removefree(lr.lr1);
    removefree(lr.lr2);  
    if (lr.lr1.free && lr.lr2.free) {
      lr.lr1=null;
      lr.lr2=null;
      lr.free=true;
      //println("rem");
    }
    else if (lr.lr1.free) {
      lr.clonefrom(lr.lr2);
    }
    else if (lr.lr2.free) {
      lr.clonefrom(lr.lr1);
    }
  }
}  
// A helper function that swaps characteristics of two lrects. Needs restruct()
void lswap(lrect lr1in, lrect lr2in) {
  lrect lr1=lr1in;
  lrect lr2=lr2in;
  lrect lrt=new lrect();
  //
  lrt.clonefrom(lr2);
  //
  lr2.clonefrom(lr1);
  //
  lr1.clonefrom(lrt);
}
