//LPacker beta

//Copyright 2008 Giannis (John) Chatzikonstantinou
//http://strukturl.blogspot.com
//http://xconst.freeshell.org

// *This file is part of LPacker*
/**
    LPacker is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    LPacker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with LPacker.  If not, see <http://www.gnu.org/licenses/>.
*/

// Recalculation function. Performs all steps necessary for the complete recalculation of item package.
void recalc() {
    // The array of items has changed in some way. Recalculate solution
      pack = new lrect(2000,Hmax,true); // 2000 is a temporary crude solution to specifying an initial space
                                        // large enough to hold all the items.
      // Update the final array with the items to be packed.
      finrects=new ArrayList();
      for (int i=0, j=initrects.size();i<j;i++) {
        lrect lr=(lrect)initrects.get(i);
        finrects.add(new lrect(lr.W,lr.H, false));
      }
      // Set the total items counter
      itemstopack=finrects.size();
      // Start packing
      lpacker();
      // Reset helper arrays and repopulate them with the new data.
      restruct();
      // The array is up to date
      change=false;
}

// A general restructuring function that performs all helper and maintenance
// subroutines on the pack, in order to keep it up to date after eg. an edit function
// or a pack function. Also calls functions to create arrays. Does not perform packing.
void restruct() {
  // Reset temporary data structures
  finrects=new ArrayList();
  caprects=new ArrayList();
  lines=new ArrayList();
  // Reset counter
  packeditems=0;
  //
  removefree(pack); // Remove the structure's free areas
  pack.updatedims(); // Update structure dimensions
  createlparts(pack,0,0,pack.W,pack.H);    // Create the temporary arrays holding the items
}

// Recursive function that creates arrays of all items to be drawn after
// processing, in order to speed up drawing and simplify interaction.
// Also allows for simpler interaction by allowing direct reference of items.
void createlparts(lrect lrin, float Xin, float Yin, float Win, float Hin) {
  lrect lr = lrin;
  lr.X=Xin;
  lr.Y=Yin;
  if (lr.leaf) { //is final(a leaf)
    // Set the items X,Y helper variables to the correct values
    finrects.add(lr);
    lrect tlr=new lrect(Xin,Yin,Win,Hin,false);
    tlr.caprear=lr.caprear;
    tlr.capfront=lr.capfront;
    caprects.add(tlr);
    packeditems++;
  }  
  else if (lr.free) {}
  // There is a tricky thing here
  // The difference between Win, Hin and lr.W, lr.H
  // Children of lrect have a dimensional analogy between them, which is
  // the ratio of the total width to the child's width in column cut, or accordingly the total height to the child's height when in row cut.
  // This ratio expresses where the separation line (between the two useful
  // spaces) will be placed. By using the same ratio, but with larger dimensions, it is possible
  // to evenly distribute excess space.
  // In this case, Win and Hin are dimensions specified initially, and the first time are usually the same
  // as the total dimensions of the pack, while lr.W and lr.H are the real dimensions of the objects.
  else if (!lr.direction) { // row cut
    float slice=lr.lr1.H/lr.H;
    createlparts(lr.lr1, Xin, Yin, Win, Hin*slice);
    createlparts(lr.lr2, Xin, Yin+Hin*slice, Win, Hin*(1-slice));
    if (!lr.lr1.free && !lr.lr2.free) {
      lines.add(new lline(Xin,Yin+Hin*slice,Xin+Win,Yin+Hin*slice,lr));
    }
  } 
  else { //column cut
    float slice=lr.lr1.W/lr.W;
    createlparts(lr.lr1, Xin, Yin, Win*slice, Hin);
    createlparts(lr.lr2, Xin+Win*slice, Yin, Win*(1-slice), Hin);
    if (!lr.lr1.free && !lr.lr2.free) {
      lines.add(new lline(Xin+Win*slice,Yin,Xin+Win*slice,Yin+Hin,lr));
    }
  }
}
//CHANGED!!!LEAST AREA
// Performs packing using various sorting methods and choosing the best result (least area);
void lpacker() {
  ArrayList finrectsw=(ArrayList)finrects.clone();
  ArrayList finrectsh=(ArrayList)finrects.clone();
  ArrayList finrectsa=(ArrayList)finrects.clone();
  ArrayList finrectsp=(ArrayList)finrects.clone();
  ArrayList finrectss=(ArrayList)finrects.clone();
  bubblesortw(finrectsw);
  bubblesorth(finrectsh);
  bubblesorta(finrectsa);
  bubblesortp(finrectsp);
  bubblesorts(finrectss);
  lrect packw=new lrect(20000,Hmax,true);
  lrect packh=new lrect(20000,Hmax,true);
  lrect packa=new lrect(20000,Hmax,true);
  lrect packp=new lrect(20000,Hmax,true);
  lrect packs=new lrect(20000,Hmax,true);
  for (int i=0,j=finrects.size();i<j;i++) {
    lrect lrw=(lrect)finrectsw.get(i);
    packw.insert(lrw);
    lrect lrh=(lrect)finrectsh.get(i);
    packh.insert(lrh);
    lrect lra=(lrect)finrectsa.get(i);
    packa.insert(lra);
    lrect lrp=(lrect)finrectsp.get(i);
    packp.insert(lrp);
    lrect lrs=(lrect)finrectss.get(i);
    packs.insert(lrs);
  }
  packw.updatedims();
  packh.updatedims();
  packa.updatedims();
  packp.updatedims();
  packs.updatedims();

  pack=packw;
  if (packh.A<pack.A) {
    pack=packh;
  }
  if (packa.A<pack.A) {
    pack=packa;
  }
  if (packp.A<pack.A) {
    pack=packp;
  }
  if (packs.A<pack.A) {
    pack=packs;
  }
}
void textrender(String title, String line1, String line2) {
  fill(255);
  noStroke();
  textFont(font2, 20);
  text(title,10,height-40);
  textFont(font, 11);
  text(line1, 10, height-25); 
    text(line2, 10, height-12); 
}
